#include <d3d11.h>
#include <d3d11_1.h>

#include "gfx/renderer_api.h"
#include "runtime/config.h"
#include "runtime/handles.h"
#include "runtime/threading.h"
#include "runtime/threading_helpers.hpp"

#include "device_objects.hpp"
#include "gpu.hpp"

RT_CVAR_I(rt_Dx11MaxCommandBuffers,
          "Maximum number of simultaneously created command buffers. Default: 1024",
          1024);

static rt_command_buffer *_buffers;
static rt_command_buffer *_first_free;
static rt_mutex *_lock;

rt_result InitCommandBufferManagement() {
    _buffers = reinterpret_cast<rt_command_buffer *>(
        calloc((size_t)rt_Dx11MaxCommandBuffers.i, sizeof(rt_command_buffer)));
    if (!_buffers)
        return RT_OUT_OF_MEMORY;
    _first_free = &_buffers[1];

    _lock = rtCreateMutex();
    if (!_lock) {
        free(_buffers);
        return RT_UNKNOWN_ERROR;
    }

    for (int i = 0; i < rt_Dx11MaxCommandBuffers.i - 1; ++i) {
        _buffers[i].next_free = &_buffers[i + 1];
    }
    return RT_SUCCESS;
}

void ShutdownCommandBufferManagement() {
    for (int i = 0; i < rt_Dx11MaxCommandBuffers.i; ++i) {
        if (_buffers[i].context)
            _buffers[i].context->Release();
    }
    free(_buffers);
    _buffers = nullptr;
}

extern "C" rt_result
RT_RENDERER_API_FN(AllocCommandBuffers)(uint32_t count,
                                        const rt_alloc_command_buffer_info *,
                                        rt_command_buffer_handle *p_command_buffers) {
    for (uint32_t i = 0; i < count; ++i) {
        rtLockMutex(_lock);
        rt_command_buffer *slot = _first_free;
        if (slot)
            _first_free = slot->next_free;
        rtUnlockMutex(_lock);

        if (!slot) {
            rtLog("dx11", "Failed to allocate a command buffer slot.");
            rtLockMutex(_lock);
            for (uint32_t j = 0; j < i; ++j) {
                rt_command_buffer *s = &_buffers[p_command_buffers[j].index];
                s->next_free         = _first_free;
                _first_free          = s;
            }
            rtUnlockMutex(_lock);
            return RT_OUT_OF_MEMORY;
        }

        if (!slot->context) {
            if (FAILED(g_gpu.device->CreateDeferredContext1(0, &slot->context))) {
                rtLog("dx11", "Failed to create a deferred context.");
                auto lock_guard = rtAutoLock(_lock);
                for (uint32_t j = 0; j < i; ++j) {
                    rt_command_buffer *s = &_buffers[p_command_buffers[j].index];
                    s->next_free         = _first_free;
                    _first_free          = s;
                }
                return RT_UNKNOWN_ERROR;
            }

#ifdef RT_DEBUG
            if (FAILED(slot->context->QueryInterface(IID_PPV_ARGS(&slot->annotation)))) {
                rtLog("dx11", "Failed to retrieve the annotation interface.");
                slot->annotation = nullptr;
            }
#endif
        } else {
            slot->context->ClearState();
        }

        slot->version                = (slot->version + 1) % RT_RENDER_BACKEND_HANDLE_MAX_VERSION;
        const uint32_t index         = (uint32_t)(slot - _buffers);
        p_command_buffers[i].version = slot->version;
        p_command_buffers[i].index   = index;
    }

    return RT_SUCCESS;
}

extern "C" rt_result
RT_RENDERER_API_FN(SubmitCommandBuffers)(rt_gpu_queue, const rt_submit_command_buffers_info *info) {
    // TODO: Handle semaphores

    // Submit the command lists to the gpu
    for (uint32_t i = 0; i < info->command_buffer_count; ++i) {
        rt_command_buffer *cmdbuf = &_buffers[info->command_buffers[i].index];
        if (cmdbuf->version != info->command_buffers[i].version) {
            rtLog("dx11", "Tried to submit an invalid command buffer (version mismatch)");
            return RT_INVALID_VALUE;
        }

        ID3D11CommandList *cmdlist;
        if (FAILED(cmdbuf->context->FinishCommandList(FALSE, &cmdlist))) {
            rtLog("dx11", "FinishCommandList failed");
            return RT_UNKNOWN_ERROR;
        }

        rtLockMutex(g_gpu.context_lock);
        g_gpu.device_context->ExecuteCommandList(cmdlist, FALSE);
        rtUnlockMutex(g_gpu.context_lock);

        rtLockMutex(_lock);
        cmdbuf->next_free = _first_free;
        _first_free       = cmdbuf;
        rtUnlockMutex(_lock);
    }
    return RT_SUCCESS;
}

rt_command_buffer *rtGetCommandBuffer(rt_command_buffer_handle handle) {
    if (!RT_IS_HANDLE_VALID(handle) || (int)handle.index >= rt_Dx11MaxCommandBuffers.i)
        return nullptr;
    auto lg = rtAutoLock(_lock);
    if (handle.version != _buffers[handle.index].version)
        return nullptr;
    return &_buffers[handle.index];
}